import os
import uuid
from datetime import datetime

import pytest

from kisters.network_store.client.network import Network
from kisters.network_store.model_library.base import LocationSet
from kisters.network_store.model_library.water import links, nodes
from kisters.water.rest_client import RESTClient
from kisters.water.rest_client.auth import OpenIDConnect

test_links = [
    links.Channel(
        uid="channel",
        source_uid="junction",
        target_uid="storage",
        length=100.0,
        created="2019-06-27T16:53:05Z",
    ),
    links.Delay(
        uid="delay",
        source_uid="junction",
        target_uid="storage",
        transit_time=10.0,
        created="2019-06-27T16:53:05Z",
    ),
    links.FlowControlledStructure(
        uid="flow_controlled_structure",
        source_uid="junction",
        target_uid="storage",
        min_flow=-1.0,
        max_flow=1.0,
        created="2019-06-27T16:53:05Z",
    ),
    links.Pipe(
        uid="pipe",
        source_uid="junction",
        target_uid="storage",
        diameter=1.0,
        length=10.0,
        roughness=10.0,
        model="hazen-williams",
        created="2019-06-27T16:53:05Z",
    ),
    links.Valve(
        uid="valve",
        source_uid="junction",
        target_uid="storage",
        diameter=10.0,
        model="prv",
        coefficient=1.0,
        setting=0.0,
        created="2019-06-27T16:53:05Z",
    ),
]
test_nodes = [
    nodes.FlowBoundary(
        uid="flow_boundary",
        location={"x": 0.0, "y": 0.0, "z": 0.0},
        created="2019-06-27T16:53:05Z",
    ),
    nodes.Junction(
        uid="junction",
        location={"x": 0.0, "y": 1.0, "z": 0.0},
        created="2019-06-27T16:53:05Z",
    ),
    nodes.LevelBoundary(
        uid="level_boundary",
        location={"x": 1.0, "y": 0.0, "z": 0.0},
        created="2019-06-27T16:53:05Z",
    ),
    nodes.Storage(
        uid="storage",
        location={"x": 1.0, "y": 1.0, "z": 0.0},
        level_volume=[{"level": 0.0, "volume": 0.0}, {"level": 10.0, "volume": 10.0}],
        created="2019-06-27T16:53:05Z",
    ),
]

rest_client = RESTClient(
    url=os.environ["NETWORK_STORE_URL"],
    authentication=OpenIDConnect(
        client_id=os.environ["CLIENT_ID"], client_secret=os.environ["CLIENT_SECRET"]
    ),
)

try:
    rest_client.get("rest/networks")
except Exception:
    pytestmark = pytest.mark.skip(
        "Currently no publicly available network store service"
    )


@pytest.fixture
def empty_network():
    network_name = f"test_{str(uuid.uuid4())[:8]}"
    network = Network(network_name, rest_client, drop_existing=True)
    yield network
    network.drop(purge=True)


@pytest.fixture
def network(empty_network):
    empty_network.initialize(nodes=test_nodes, links=test_links)
    return empty_network  # no longer empty


def test_initialize(empty_network):
    empty_network.initialize(nodes=test_nodes, links=test_links)


def test_no_auth():
    unauth_client = RESTClient(url=os.environ["NETWORK_STORE_URL"])
    with pytest.raises(Exception):
        unauth_client.get(("rest", "networks"))


def test_get_extents(network):
    extent = network.get_extent()
    assert extent.dict() == {"x": (0, 1), "y": (0, 1), "z": (0, 0)}
    schematic_extent = network.get_extent(location_set=LocationSet.SCHEMATIC)
    assert schematic_extent.dict() == {"x": (0, 1), "y": (0, 1), "z": (0, 0)}


def test_get_nodes(network):
    remote = network.get_nodes()
    remote = sorted(remote, key=lambda e: e.uid)
    assert test_nodes == remote


def test_get_links(network):
    remote = network.get_links()
    remote = sorted(remote, key=lambda e: e.uid)
    assert test_links == remote


def test_get_nodes_filtered(network):
    remote = network.get_nodes(dt=datetime(2019, 6, 27, 16, 53, 4))
    assert remote == []
    remote = network.get_nodes(dt=datetime(2019, 6, 27, 16, 53, 5))
    remote = sorted(remote, key=lambda e: e.uid)
    assert test_nodes == remote

    subset_nodes = test_nodes[:2]
    remote = network.get_nodes(uids=[e.uid for e in subset_nodes])
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_nodes

    subset_nodes = test_nodes[:2]
    remote = network.get_nodes(display_names=[e.display_name for e in subset_nodes])
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_nodes

    subset_node = test_nodes[0]
    remote = network.get_nodes(element_class=subset_node.element_class)
    assert remote == [subset_node]

    subset_node = test_nodes[0]
    remote = network.get_nodes(element_class=type(subset_node))
    assert remote == [subset_node]

    subset_nodes = test_nodes[:2]
    remote = network.get_nodes(extent={"x": [-0.5, 0.5]})
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_nodes

    subset_nodes = test_nodes[:2]
    remote = network.get_nodes(schematic_extent={"x": [-0.5, 0.5]})
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_nodes

    remote = network.get_nodes(extent={"z": [-0.5, 0.5]})
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_nodes

    remote = network.get_nodes(schematic_extent={"z": [-0.5, 0.5]})
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_nodes


def test_get_links_filtered(network):
    remote = network.get_links(dt=datetime(2019, 6, 27, 16, 53, 4))
    assert remote == []
    remote = network.get_links(dt=datetime(2019, 6, 27, 16, 53, 5))
    remote = sorted(remote, key=lambda e: e.uid)
    assert test_links == remote

    subset_links = test_links[:2]
    remote = network.get_links(uids=[e.uid for e in subset_links])
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_links

    subset_links = test_links[:2]
    remote = network.get_links(display_names=[e.display_name for e in subset_links])
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == subset_links

    subset_link = test_links[0]
    remote = network.get_links(element_class=subset_link.element_class)
    assert remote == [subset_link]

    subset_link = test_links[0]
    remote = network.get_links(element_class=type(subset_link))
    assert remote == [subset_link]

    remote = network.get_links(adjacent_nodes=["junction", "storage"])
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_links

    remote = network.get_links(adjacent_nodes=["junction"])
    assert remote == []

    remote = network.get_links(adjacent_nodes=["junction"], only_interior=False)
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_links


def test_drop_nodes(network):
    network.drop_nodes([node.uid for node in test_nodes])
    nodes = network.get_nodes()
    assert nodes == []


def test_drop_links(network):
    network.drop_links([node.uid for node in test_links])
    links = network.get_links()
    assert links == []


def test_save_nodes(empty_network):
    empty_network.save_nodes(test_nodes)
    remote = empty_network.get_nodes()
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_nodes


def test_save_links(empty_network):
    empty_network.save_links(test_links)
    remote = empty_network.get_links()
    remote = sorted(remote, key=lambda e: e.uid)
    assert remote == test_links


def test_ValueError(network):
    with pytest.raises(ValueError):
        network.save_nodes(test_links)
    with pytest.raises(ValueError):
        network.save_links(test_nodes)


def test_drop(network):
    network.drop()
    links = network.get_links()
    assert links == []
    nodes = network.get_nodes()
    assert nodes == []



from kisters.water.rest_client import RESTClient

from ..network import Network


class NetworkStore:
    _resource = "rest", "networks"

    def __init__(self, client: RESTClient):
        self._client = client

    def list_networks(self):
        return self._client.get(self._resource)

    def get_network(self, network_uid):
        return Network(network_uid, self._client)



from datetime import datetime
from typing import List, Optional

from kisters.network_store.model_library.base import (
    BaseGroup,
    BaseLink,
    BaseNode,
    LocationExtent,
    LocationSet,
)
from kisters.network_store.model_library.util import element_from_dict
from kisters.water.rest_client import RESTClient


class Network:
    """Access the network store API

    Provideds convenient methods to access the network store API from python
    scripts and adapters.

    :param name:
        ID of the network to use
    :param client:
        An instance of RESTClient or equivalent
    :param drop_existing:
        Delete any existing elements (deprecated- use `network.drop()`)
    :param dt:
        Optional default dt that will be used for all queries
    """

    _resource = "rest", "networks"

    def __init__(
        self,
        name: str,
        client: RESTClient,
        drop_existing: bool = False,  # Deprecated- use `network.drop()`
        dt: Optional[datetime] = None,
        group_uid: Optional[str] = None,
    ):
        self._name = name
        self._client = client
        self._network_resource = (*self._resource, self.name)
        if drop_existing:
            self.drop()
        self._default_dt = dt
        self._default_group_uid = group_uid

    def initialize(
        self,
        nodes: Optional[List[BaseNode]] = None,
        links: Optional[List[BaseLink]] = None,
        groups: Optional[List[BaseGroup]] = None,
    ):
        """Clear a network, optionally setting new links, nodes, and groups

        Note that this does not purge the network history- older versions of
        networks can be revisited even after being initialised. Useful for
        defining a complete network in one call.

        :param nodes:
            Optional list of nodes
        :param links:
            Optional list of links
        :param groups:
            Optional list of groups
        """
        nodes = ",".join((e.json(exclude_none=True) for e in nodes)) if nodes else ""
        links = ",".join((e.json(exclude_none=True) for e in links)) if links else ""
        groups = ",".join((e.json(exclude_none=True) for e in groups)) if groups else ""
        network_data = f'{{"nodes":[{nodes}],"links":[{links}],"groups":[{groups}]}}'
        self._client.post(
            self._network_resource,
            data=network_data,
            headers={"content-type": "application/json"},
        )

    def get_extent(
        self,
        location_set: LocationSet = LocationSet.GEOGRAPHIC,
        *,
        dt: Optional[datetime] = None,
    ) -> LocationExtent:
        """Gets the min and max values over all node locations

        Note that at the moment, only nodes are considered. Link vertices are
        ignored.

        :param location_set:
            The name of the location attribute, one of the LocationSet enum
        :param dt:
            Optional timestamp to access a historical version of the model
        """
        params = {"location_set": location_set}
        dt = dt or self._default_dt
        if dt:
            params["datetime"] = dt.isoformat()
        extents = self._client.get((*self._network_resource, "extent"), params=params)
        return LocationExtent.parse_obj(extents)

    def get_links(
        self,
        uids: Optional[List[str]] = None,
        display_names: Optional[List[str]] = None,
        element_class: Optional[str] = None,
        group_uids: Optional[List[str]] = None,
        adjacent_nodes: Optional[List[str]] = None,
        only_interior: Optional[bool] = True,
        include_subgroups: Optional[bool] = False,
        dt: Optional[datetime] = None,
    ) -> List[BaseLink]:
        """Gets an iterable of links

        Gets the the links in the network. The links are filterable by optional
        kwargs.

        :param uids:
            Optional list of uid strings to match
        :param display_names:
            Optional list of display name strings to match
        :param element_class:
            Optional element class string to match
        :param adjacent_nodes:
            Optional list of node uids that the links are connected to
        :param only_interior:
            Don't to match links that are attached at only one end to an adjacent_node
        :param include_subgroups:
            Recursively include groups that are subgroups of group_uids
        :param dt:
            Optional timestamp to access a historical version of the model
        """
        params = {}
        data = {}
        if uids:
            data["uids"] = uids
        if display_names:
            data["display_names"] = display_names
        if element_class:
            if isinstance(element_class, str):
                params["element_class"] = element_class
            elif hasattr(element_class, "__name__"):
                params["element_class"] = element_class.__name__
            else:
                raise ValueError(
                    "kwarg element_class {} should be string or element class".format(
                        element_class
                    )
                )
        if group_uids:
            data["group_uids"] = group_uids
        elif self._default_group_uid:
            data["group_uids"] = [self._default_group_uid]
        if adjacent_nodes:
            data["adjacent_node_uids"] = [
                (node if isinstance(node, str) else node.uid) for node in adjacent_nodes
            ]
        if not only_interior:
            params["only_interior"] = "False"
        if include_subgroups:
            params["include_subgroups"] = include_subgroups
        dt = dt or self._default_dt
        if dt:
            params["datetime"] = dt.isoformat()
        result = self._client.post(
            (*self._network_resource, "links", "search"),
            params=params,
            json=data,
        )
        return [element_from_dict(e) for e in result]

    def get_nodes(
        self,
        uids: Optional[List[str]] = None,
        display_names: Optional[List[str]] = None,
        element_class: Optional[List[str]] = None,
        group_uids: Optional[List[str]] = None,
        extent: Optional[LocationExtent] = None,
        schematic_extent: Optional[LocationExtent] = None,
        include_subgroups: Optional[bool] = False,
        dt: Optional[datetime] = None,
    ) -> List[BaseNode]:
        """Gets an iterable of nodes

        Gets the the nodes in the network. The nodes are filterable by optional
        kwargs.

        :param uids:
            Optional list of uid strings to match
        :param display_names:
            Optional list of display name strings to match
        :param element_class:
            Optional element type string to match
        :param extent:
            Optional mapping of extent dimensions to min and max extent of that
            dimension that returned nodes should be found within.
        :param schematic_extent:
            Optional mapping of schematic_extent dimensions to min and max extent
            of that dimension that returned nodes should be found within.
        :param include_subgroups:
            Recursively include groups that are subgroups of group_uids
        :param dt:
            Optional timestamp to access a historical version of the model
        """
        params = {}
        data = {}
        if uids:
            data["uids"] = uids
        if display_names:
            data["display_names"] = display_names
        if element_class:
            if isinstance(element_class, str):
                params["element_class"] = element_class
            elif hasattr(element_class, "__name__"):
                params["element_class"] = element_class.__name__
            else:
                raise ValueError(
                    "kwarg element_class {} should be string or element class".format(
                        element_class
                    )
                )
        if group_uids:
            data["group_uids"] = group_uids
        elif self._default_group_uid:
            data["group_uids"] = [self._default_group_uid]
        if extent:
            if isinstance(extent, dict):
                extent = LocationExtent.parse_obj(extent)
            data["extent"] = extent.dict()
        if schematic_extent:
            if isinstance(schematic_extent, dict):
                schematic_extent = LocationExtent.parse_obj(schematic_extent)
            data["schematic_extent"] = schematic_extent.dict()
        if include_subgroups:
            params["include_subgroups"] = include_subgroups
        dt = dt or self._default_dt
        if dt:
            params["datetime"] = dt.isoformat()

        result = self._client.post(
            (*self._network_resource, "nodes", "search"),
            params=params,
            json=data,
        )
        return [element_from_dict(e) for e in result]

    def get_groups(
        self,
        uids: Optional[List[str]] = None,
        display_names: Optional[List[str]] = None,
        element_class: Optional[List[str]] = None,
        group_uids: Optional[List[str]] = None,
        include_subgroups: Optional[bool] = False,
        dt: Optional[datetime] = None,
    ) -> List[BaseGroup]:
        """Gets an iterable of groups

        Gets the the groups in the network. The groups are filterable by optional
        kwargs.

        :param uids:
            Optional list of uid strings to match
        :param display_names:
            Optional list of display name strings to match
        :param element_class:
            Optional element type string to match
        :param include_subgroups:
            Recursively include groups that are subgroups of group_uids
        :param dt:
            Optional timestamp to access a historical version of the model
        """
        params = {}
        data = {}
        if uids:
            data["uids"] = uids
        if display_names:
            data["display_names"] = display_names
        if element_class:
            if isinstance(element_class, str):
                params["element_class"] = element_class
            elif hasattr(element_class, "__name__"):
                params["element_class"] = element_class.__name__
            else:
                raise ValueError(
                    "kwarg element_class {} should be string or element class".format(
                        element_class
                    )
                )
        if group_uids:
            data["group_uids"] = group_uids
        elif self._default_group_uid:
            data["group_uids"] = [self._default_group_uid]
        if include_subgroups:
            params["include_subgroups"] = include_subgroups
        dt = dt or self._default_dt
        if dt:
            params["datetime"] = dt.isoformat()

        result = self._client.post(
            (*self._network_resource, "groups", "search"),
            params=params,
            json=data,
        )
        return [element_from_dict(e) for e in result]

    def save_nodes(self, nodes: List[BaseNode]):
        """Save a list of nodes"""
        if not nodes:
            return
        nodes = ",".join((e.json(exclude_none=True) for e in nodes))
        nodes_data = f'{{"elements":[{nodes}]}}'
        self._client.post(
            (*self._network_resource, "nodes"),
            data=nodes_data,
            headers={"content-type": "application/json"},
        )

    def save_links(self, links: List[BaseLink]):
        """Save a list of links"""
        if not links:
            return
        links = ",".join((e.json(exclude_none=True) for e in links))
        links_data = f'{{"elements":[{links}]}}'
        self._client.post(
            (*self._network_resource, "links"),
            data=links_data,
            headers={"content-type": "application/json"},
        )

    def save_groups(self, groups: List[BaseGroup]):
        """Save a list of groups"""
        if not groups:
            return
        groups = ",".join((e.json(exclude_none=True) for e in groups))
        groups_data = f'{{"elements":[{groups}]}}'
        self._client.post(
            (*self._network_resource, "groups"),
            data=groups_data,
            headers={"content-type": "application/json"},
        )

    def drop(self, purge: bool = False):
        """Removes the network"""
        self._client.delete(self._network_resource, params={"purge": purge})

    def drop_links(self, uids: Optional[List[str]] = None, purge: bool = False):
        """Deletes the links with the given uids"""
        self._client.delete(
            (*self._network_resource, "links"),
            json={"uids": uids} if uids else None,
            params={"purge": purge},
        )

    def drop_nodes(self, uids: Optional[List[str]] = None, purge: bool = False):
        """Deletes the nodes with the given uids"""
        self._client.delete(
            (*self._network_resource, "nodes"),
            json={"uids": uids} if uids else None,
            params={"purge": purge},
        )

    def drop_groups(self, uids: Optional[List[str]] = None, purge: bool = False):
        """Deletes the groups with the given uids"""
        self._client.delete(
            (*self._network_resource, "groups"),
            json={"uids": uids} if uids else None,
            params={"purge": purge},
        )

    @property
    def name(self) -> str:
        """The name of the network"""
        return self._name



import os

import pytest

from kisters.network_store.client.network_store import NetworkStore
from kisters.water.rest_client import RESTClient
from kisters.water.rest_client.auth import OpenIDConnect

rest_client = RESTClient(
    url=os.environ["NETWORK_STORE_URL"],
    authentication=OpenIDConnect(
        client_id=os.environ["CLIENT_ID"], client_secret=os.environ["CLIENT_SECRET"]
    ),
)

try:
    rest_client.get("rest/networks")
except Exception:
    pytestmark = pytest.mark.skip(
        "Currently no publicly available network store service"
    )


def test_store():
    store = NetworkStore(rest_client)
    store.list_networks()
    network = store.get_network("tmp")
    network.drop()



from os import path

from setuptools import find_namespace_packages, setup

# Get the long description from the README file
here = path.abspath(path.dirname(__file__))
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="kisters.network_store.client.network",
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
    author="Jesse VanderWees",
    author_email="jesse.vanderwees@kisters-bv.nl",
    description="Client library for the Kisters Network Store service",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://gitlab.com/kisters/network-store/client",
    packages=find_namespace_packages(include=["kisters.*"]),
    zip_safe=False,
    install_requires=[
        "kisters.network_store.model_library>=0.2.6",
        "kisters.water.rest_client>=0.0.8",
    ],
    extras_require={
        "test": ["kisters.network_store.model_library.water>=0.2.8", "pytest"],
        "water": ["kisters.network_store.model_library.water>=0.2.8"],
    },
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Operating System :: OS Independent",
    ],
)
